import os
import json
import yaml
import inspect
from typing import Union, Any

import torch

from lume_model.base import parse_config, recursive_serialize
from lume_model.models.torch_model import TorchModel
from lume_model.mlflow_utils import register_model


class TorchModule(torch.nn.Module):
    """Wrapper to allow a LUME TorchModel to be used like a torch.nn.Module.

    As the base model within the TorchModel is assumed to be fixed during instantiation,
    so is the TorchModule.
    """

    def __init__(
        self,
        *args,
        model: TorchModel = None,
        input_order: list[str] = None,
        output_order: list[str] = None,
    ):
        """Initializes TorchModule.

        Args:
            *args: Accepts a single argument which is the model configuration as dictionary, YAML or JSON
              formatted string or file path.

        Keyword Args:
            model: The TorchModel instance to wrap around. If config is None, this has to be defined.
            input_order: Input names in the order they are passed to the model. If None, the input order of the
              TorchModel is used.
            output_order: Output names in the order they are returned by the model. If None, the output order of
              the TorchModel is used.
        """
        if all(arg is None for arg in [*args, model]):
            raise ValueError(
                "Either a YAML string has to be given or model has to be defined."
            )
        super().__init__()
        if len(args) == 1:
            if not all(v is None for v in [model, input_order, output_order]):
                raise ValueError(
                    "Cannot specify YAML string and keyword arguments for TorchModule init."
                )
            model_fields = {f"model.{k}": v for k, v in TorchModel.model_fields.items()}
            kwargs = parse_config(args[0], model_fields)
            kwargs["model"] = TorchModel(kwargs["model"])
            self.__init__(**kwargs)
        elif len(args) > 1:
            raise ValueError(
                "Arguments to TorchModule must be either a single YAML string or keyword arguments."
            )
        else:
            self._model = model
            self._input_order = input_order
            self._output_order = output_order
            self.register_module("base_model", self._model.model)
            for i, input_transformer in enumerate(self._model.input_transformers):
                self.register_module(f"input_transformers_{i}", input_transformer)
            for i, output_transformer in enumerate(self._model.output_transformers):
                self.register_module(f"output_transformers_{i}", output_transformer)
            if not model.model.training:  # TorchModel defines train/eval mode
                self.eval()

    @property
    def model(self):
        return self._model

    @property
    def input_order(self):
        if self._input_order is None:
            return self._model.input_names
        else:
            return self._input_order

    @property
    def output_order(self):
        if self._output_order is None:
            return self._model.output_names
        else:
            return self._output_order

    def forward(self, x: torch.Tensor):
        # input shape: [n_batch, n_samples, n_dim]
        x = self._validate_input(x)
        model_input = self._tensor_to_dictionary(x)
        y_model = self.evaluate_model(model_input)
        y_model = self.manipulate_output(y_model)
        # squeeze for use as prior mean in botorch GPs
        y = self._dictionary_to_tensor(y_model).squeeze()
        return y

    def yaml(
        self,
        base_key: str = "",
        file_prefix: str = "",
        save_models: bool = False,
        save_jit: bool = False,
    ) -> str:
        """Serializes the object and returns a YAML formatted string defining the TorchModule instance.

        Args:
            base_key: Base key for serialization.
            file_prefix: Prefix for generated filenames.
            save_models: Determines whether models are saved to file.
            save_jit: Determines whether the structure of the model is saved as TorchScript

        Returns:
            YAML formatted string defining the TorchModule instance.
        """
        d = {}
        for k, v in inspect.signature(TorchModule.__init__).parameters.items():
            if k not in ["self", "args", "model"]:
                d[k] = getattr(self, k)
        output = json.loads(
            json.dumps(
                recursive_serialize(d, base_key, file_prefix, save_models, save_jit)
            )
        )
        model_output = json.loads(
            self._model.to_json(
                base_key=base_key,
                file_prefix=file_prefix,
                save_models=save_models,
                save_jit=save_jit,
            )
        )
        output["model"] = model_output
        # create YAML formatted string
        s = yaml.dump(
            {"model_class": self.__class__.__name__} | output,
            default_flow_style=None,
            sort_keys=False,
        )
        return s

    def dump(
        self,
        file: Union[str, os.PathLike],
        save_models: bool = True,
        base_key: str = "",
        save_jit: bool = False,
    ):
        """Returns and optionally saves YAML formatted string defining the model.

        Args:
            file: File path to which the YAML formatted string and corresponding files are saved.
            base_key: Base key for serialization.
            save_models: Determines whether models are saved to file.
            save_jit : Whether the model is saved using just in time pytorch method
        """
        file_prefix = os.path.splitext(file)[0]
        with open(file, "w") as f:
            f.write(
                self.yaml(
                    save_models=save_models,
                    base_key=base_key,
                    file_prefix=file_prefix,
                    save_jit=save_jit,
                )
            )

    def evaluate_model(self, x: dict[str, torch.Tensor]):
        """Placeholder method to modify model calls."""
        return self._model.evaluate(x)

    def manipulate_output(self, y_model: dict[str, torch.Tensor]):
        """Placeholder method to modify the model output."""
        return y_model

    def _tensor_to_dictionary(self, x: torch.Tensor):
        input_dict = {}
        for idx, input_name in enumerate(self.input_order):
            input_dict[input_name] = x[..., idx].unsqueeze(-1)
        return input_dict

    def _dictionary_to_tensor(self, y_model: dict[str, torch.Tensor]):
        output_tensor = torch.stack(
            [y_model[output_name].unsqueeze(-1) for output_name in self.output_order],
            dim=-1,
        )
        return output_tensor

    @staticmethod
    def _validate_input(x: torch.Tensor) -> torch.Tensor:
        if x.dim() <= 1:
            raise ValueError(
                f"Expected input dim to be at least 2 ([n_samples, n_features]), received: {tuple(x.shape)}"
            )
        else:
            return x

    def register_to_mlflow(
        self,
        input: torch.Tensor,
        artifact_path: str,
        registered_model_name: str | None = None,
        tags: dict[str, Any] | None = None,
        version_tags: dict[str, Any] | None = None,
        alias: str | None = None,
        run_name: str | None = None,
        log_model_dump: bool = True,
        save_jit: bool = False,
        **kwargs,
    ):
        """
        Registers the model to MLflow if mlflow is installed. Each time this function is called, a new version
        of the model is created. The model is saved to the tracking server or local directory, depending on the
        MLFLOW_TRACKING_URI.

        If no tracking server is set up, data and artifacts are saved directly under your current directory. To set up
        a tracking server, set the environment variable MLFLOW_TRACKING_URI, e.g. a local port/path. See
        https://mlflow.org/docs/latest/getting-started/intro-quickstart/ for more info.

        Args:
            input: Input tensor to infer the model signature.
            artifact_path: Path to store the model in MLflow.
            registered_model_name: Name of the registered model in MLflow. Optional.
            tags: Tags to add to the MLflow model. Optional.
            version_tags: Tags to add to this MLflow model version. Optional.
            alias: Alias to add to this MLflow model version. Optional.
            run_name: Name of the MLflow run. Optional.
            log_model_dump: Whether to log the model dump files as artifacts. Optional.
            save_jit: Whether to save the model as TorchScript when calling model.dump, if log_model_dump=True. Optional.
            **kwargs: Additional arguments for mlflow.pyfunc.log_model.

        Returns:
            Model info metadata, mlflow.models.model.ModelInfo.
        """
        return register_model(
            self,
            input,
            artifact_path,
            registered_model_name,
            tags,
            version_tags,
            alias,
            run_name,
            log_model_dump,
            save_jit,
            **kwargs,
        )



import os
import pytest
import random
from copy import deepcopy

try:
    import torch
    from botorch.models import SingleTaskGP, MultiTaskGP, ModelListGP

    torch.manual_seed(42)
except ImportError:
    pass

from lume_model.models.gp_model import GPModel
from lume_model.variables import ScalarVariable, DistributionVariable

random.seed(42)


class TestGPModel:
    def test_validates_model_type(self, single_task_gp_model_kwargs):
        with pytest.raises(OSError):
            GPModel(
                model="invalid_model_type",
                input_variables=single_task_gp_model_kwargs["input_variables"],
                output_variables=single_task_gp_model_kwargs["output_variables"],
            )

    def test_raises_error_for_missing_transformer_file(
        self, single_task_gp_model_kwargs
    ):
        with pytest.raises(OSError):
            GPModel(
                model=SingleTaskGP(
                    torch.rand(10, 1, dtype=torch.double),
                    torch.rand(10, 1, dtype=torch.double),
                ),
                input_transformers=["non_existent_transformer.pt"],
                output_transformers=["non_existent_transformer.pt"],
                input_variables=single_task_gp_model_kwargs["input_variables"],
                output_variables=single_task_gp_model_kwargs["output_variables"],
            )

    def test_single_task_gp(self, single_task_gp_model_kwargs):
        gp_model = GPModel(
            model=single_task_gp_model_kwargs["model"],
            input_variables=single_task_gp_model_kwargs["input_variables"],
            output_variables=single_task_gp_model_kwargs["output_variables"],
            input_transformers=single_task_gp_model_kwargs["input_transformers"],
            output_transformers=single_task_gp_model_kwargs["output_transformers"],
        )
        assert isinstance(gp_model.model, SingleTaskGP)

        original_model = single_task_gp_model_kwargs["model"]
        input_transformer = single_task_gp_model_kwargs["input_transformers"][0]
        output_transformer = single_task_gp_model_kwargs["output_transformers"][0]

        test_x = torch.rand(10, 1).to(torch.double)
        # Predict with original model
        test_x_tf = input_transformer.transform(test_x)
        original_pred = original_model.posterior(test_x_tf)
        original_mean = output_transformer.untransform(original_pred.mean)
        original_variance = (
            abs(output_transformer.coefficient) ** 2 * original_pred.variance
        )
        # Predict with GPModel
        lume_pred = gp_model.evaluate({"input": test_x})

        assert gp_model.get_input_size() == 1
        assert gp_model.get_output_size() == 2
        assert len(lume_pred) == 2
        assert torch.allclose(original_mean[:, 0], lume_pred["output1"].mean)
        assert torch.allclose(original_variance[:, 0], lume_pred["output1"].variance)
        assert torch.allclose(original_mean[:, 1], lume_pred["output2"].mean)
        assert torch.allclose(original_variance[:, 1], lume_pred["output2"].variance)

        # Test with observation noise set to True
        original_pred = original_model.posterior(test_x_tf, observation_noise=True)
        # Output transformer is a ReversibleInputTransform type
        original_mean = output_transformer.untransform(original_pred.mean)
        original_variance = (
            abs(output_transformer.coefficient) ** 2 * original_pred.variance
        )
        # Predict with GPModel
        lume_pred = gp_model.evaluate({"input": test_x}, observation_noise=True)

        assert gp_model.get_input_size() == 1
        assert gp_model.get_output_size() == 2
        assert len(lume_pred) == 2
        assert torch.allclose(original_mean[:, 0], lume_pred["output1"].mean)
        assert torch.allclose(original_variance[:, 0], lume_pred["output1"].variance)
        assert torch.allclose(original_mean[:, 1], lume_pred["output2"].mean)
        assert torch.allclose(original_variance[:, 1], lume_pred["output2"].variance)

        # Test batched evaluation
        batch_test_x = test_x.reshape(-1, 1).repeat(2, 1, 1)
        # Predict with original model
        batch_test_x_tf = input_transformer.transform(batch_test_x)
        original_pred = original_model.posterior(batch_test_x_tf)
        # Output transformer is a ReversibleInputTransform type
        original_mean = output_transformer.untransform(original_pred.mean)
        original_variance = (
            abs(output_transformer.coefficient) ** 2 * original_pred.variance
        )
        # Predict with GPModel
        lume_pred = gp_model.evaluate({"input": batch_test_x})

        assert lume_pred["output1"].mean.shape == (2, 10) and lume_pred[
            "output1"
        ].variance.shape == (2, 10)
        assert lume_pred["output2"].mean.shape == (2, 10) and lume_pred[
            "output2"
        ].variance.shape == (2, 10)
        assert original_pred.mean.shape == (
            2,
            10,
            2,
        ) and original_pred.variance.shape == (2, 10, 2)
        assert torch.allclose(original_mean[:, :, 0], lume_pred["output1"].mean)
        assert torch.allclose(original_variance[:, :, 0], lume_pred["output1"].variance)
        assert torch.allclose(original_mean[:, :, 1], lume_pred["output2"].mean)
        assert torch.allclose(original_variance[:, :, 1], lume_pred["output2"].variance)

        # Dump and load the model
        gp_model.dump("test.yml")
        loaded_model = GPModel("test.yml")
        lume_pred_loaded = loaded_model.evaluate({"input": batch_test_x})
        assert torch.allclose(
            lume_pred["output1"].mean, lume_pred_loaded["output1"].mean
        )
        assert torch.allclose(
            lume_pred["output1"].variance, lume_pred_loaded["output1"].variance
        )
        assert torch.allclose(
            lume_pred["output2"].mean, lume_pred_loaded["output2"].mean
        )
        assert torch.allclose(
            lume_pred["output2"].variance, lume_pred_loaded["output2"].variance
        )
        os.remove("test.yml")
        os.remove("test_model.pt")
        os.remove("test_input_transformers_0.pt")
        os.remove("test_output_transformers_0.pt")

    def test_multi_task_gp(self, multi_task_gp_model_kwargs):
        gp_model = GPModel(
            model=multi_task_gp_model_kwargs["model"],
            input_variables=multi_task_gp_model_kwargs["input_variables"],
            output_variables=multi_task_gp_model_kwargs["output_variables"],
            input_transformers=multi_task_gp_model_kwargs["input_transformers"],
            output_transformers=multi_task_gp_model_kwargs["output_transformers"],
        )
        assert isinstance(gp_model.model, MultiTaskGP)

        original_model = multi_task_gp_model_kwargs["model"]
        input_transformer = multi_task_gp_model_kwargs["input_transformers"][0]
        output_transformer = multi_task_gp_model_kwargs["output_transformers"][0]

        # Make sure to use double precision for all GP tests
        test_x = torch.linspace(0, 1, 200).reshape(-1, 1).to(torch.double)
        # Predict with original model
        test_x_tf = input_transformer.transform(test_x)

        original_pred = original_model.posterior(test_x_tf)
        m = original_pred.mean
        v = original_pred.variance
        # Output transformer is a OutcomeTransform type
        original_mean = output_transformer.stdvs.squeeze(
            0
        ) * m + output_transformer.means.squeeze(0)
        original_variance = abs(output_transformer.stdvs.squeeze(0)) ** 2 * v
        # Predict with GPModel
        lume_pred = gp_model.evaluate({"input": test_x})

        assert gp_model.get_input_size() == 1
        assert gp_model.get_output_size() == 2
        assert len(lume_pred) == 2
        assert lume_pred["output1"].mean.shape == (200,) and lume_pred[
            "output1"
        ].variance.shape == (200,)
        assert lume_pred["output2"].mean.shape == (200,) and lume_pred[
            "output2"
        ].variance.shape == (200,)
        assert original_pred.mean.shape == (
            200,
            2,
        ) and original_pred.variance.shape == (200, 2)
        assert torch.allclose(original_mean[:, 0], lume_pred["output1"].mean)
        assert torch.allclose(original_variance[:, 0], lume_pred["output1"].variance)
        assert torch.allclose(original_mean[:, 1], lume_pred["output2"].mean)
        assert torch.allclose(original_variance[:, 1], lume_pred["output2"].variance)

        # Test batched evaluation
        batch_test_x = test_x.repeat(2, 1, 1)
        # Predict with original model
        batch_test_x_tf = input_transformer.transform(batch_test_x)
        original_pred = original_model.posterior(batch_test_x_tf)
        m = original_pred.mean
        v = original_pred.variance
        # Output transformer is a OutcomeTransform type
        original_mean = output_transformer.stdvs.squeeze(
            0
        ) * m + output_transformer.means.squeeze(0)
        original_variance = abs(output_transformer.stdvs.squeeze(0)) ** 2 * v
        # Predict with GPModel
        lume_pred = gp_model.evaluate({"input": batch_test_x})

        assert lume_pred["output1"].mean.shape == (2, 200) and lume_pred[
            "output1"
        ].variance.shape == (2, 200)
        assert lume_pred["output2"].mean.shape == (2, 200) and lume_pred[
            "output2"
        ].variance.shape == (2, 200)
        assert original_pred.mean.shape == (
            2,
            200,
            2,
        ) and original_pred.variance.shape == (2, 200, 2)
        assert torch.allclose(original_mean[:, :, 0], lume_pred["output1"].mean)
        assert torch.allclose(original_variance[:, :, 0], lume_pred["output1"].variance)
        assert torch.allclose(original_mean[:, :, 1], lume_pred["output2"].mean)
        assert torch.allclose(original_variance[:, :, 1], lume_pred["output2"].variance)

        # Dump and load the model
        gp_model.dump("test.yml")
        loaded_model = GPModel("test.yml")
        lume_pred_loaded = loaded_model.evaluate({"input": batch_test_x})
        assert torch.allclose(
            lume_pred["output1"].mean, lume_pred_loaded["output1"].mean
        )
        assert torch.allclose(
            lume_pred["output1"].variance, lume_pred_loaded["output1"].variance
        )
        assert torch.allclose(
            lume_pred["output2"].mean, lume_pred_loaded["output2"].mean
        )
        assert torch.allclose(
            lume_pred["output2"].variance, lume_pred_loaded["output2"].variance
        )
        os.remove("test.yml")
        os.remove("test_model.pt")
        os.remove("test_input_transformers_0.pt")
        os.remove("test_output_transformers_0.pt")

    def test_model_list_gp(self, create_single_task_gp, create_multi_task_gp):
        # based on botorch/test/models/test_model_list_gp_regression.py

        # SingleTaskGP
        model1, test_x = create_single_task_gp
        model_list_gp = ModelListGP(model1)
        posterior = model_list_gp.posterior(test_x)

        input_variables = [ScalarVariable(name="input")]
        output_variables = [DistributionVariable(name="output")]
        gp_lume_model = GPModel(
            model=model_list_gp,
            input_variables=input_variables,
            output_variables=output_variables,
        )
        lume_pred = gp_lume_model.evaluate({"input": test_x})
        assert gp_lume_model.get_input_size() == 1
        assert gp_lume_model.get_output_size() == 1
        assert len(lume_pred) == 1
        assert torch.allclose(posterior.mean, lume_pred["output"].mean.unsqueeze(-1))
        assert torch.allclose(
            posterior.variance, lume_pred["output"].variance.unsqueeze(-1)
        )

        # MultiTaskGP
        model, model2, train_x_raw = create_multi_task_gp

        # Wrap a single single-output MTGP.
        model_list_gp = ModelListGP(model)
        with torch.no_grad():
            posterior = model_list_gp.posterior(train_x_raw)
        input_variables = [ScalarVariable(name="input1")]
        output_variables = [DistributionVariable(name="output1")]
        gp_lume_model = GPModel(
            model=model_list_gp,
            input_variables=input_variables,
            output_variables=output_variables,
        )
        input_dict = {"input1": train_x_raw}
        output_dict = gp_lume_model.evaluate(input_dict)
        assert gp_lume_model.get_input_size() == 1
        assert gp_lume_model.get_output_size() == 1
        assert len(output_dict) == 1
        assert torch.allclose(output_dict["output1"].mean, posterior.mean.squeeze(-1))
        assert torch.allclose(
            output_dict["output1"].variance, posterior.variance.squeeze(-1)
        )

        # Wrap two single-output MTGPs.
        model_list_gp = ModelListGP(model, model)
        with torch.no_grad():
            posterior = model_list_gp.posterior(train_x_raw)
        output_variables = [
            DistributionVariable(name="output1"),
            DistributionVariable(name="output2"),
        ]
        gp_lume_model = GPModel(
            model=model_list_gp,
            input_variables=input_variables,
            output_variables=output_variables,
        )
        output_dict = gp_lume_model.evaluate(input_dict)
        assert gp_lume_model.get_input_size() == 1
        assert gp_lume_model.get_output_size() == 2
        assert len(output_dict) == 2
        assert torch.allclose(output_dict["output1"].mean, posterior.mean[:, 0])
        assert torch.allclose(output_dict["output1"].variance, posterior.variance[:, 0])
        assert torch.allclose(output_dict["output2"].mean, posterior.mean[:, 1])
        assert torch.allclose(output_dict["output2"].variance, posterior.variance[:, 1])

        # Wrap a multi-output MTGP.
        model_list_gp = ModelListGP(model2)
        with torch.no_grad():
            posterior = model_list_gp.posterior(train_x_raw)
        gp_lume_model = GPModel(
            model=model_list_gp,
            input_variables=input_variables,
            output_variables=output_variables,
        )
        output_dict = gp_lume_model.evaluate(input_dict)
        assert gp_lume_model.get_input_size() == 1
        assert gp_lume_model.get_output_size() == 2
        assert len(output_dict) == 2
        assert torch.allclose(output_dict["output1"].mean, posterior.mean[:, 0])
        assert torch.allclose(output_dict["output1"].variance, posterior.variance[:, 0])
        assert torch.allclose(output_dict["output2"].mean, posterior.mean[:, 1])
        assert torch.allclose(output_dict["output2"].variance, posterior.variance[:, 1])

        # Mix of multi-output and single-output MTGPs.
        model_list_gp = ModelListGP(model, model2, deepcopy(model))
        with torch.no_grad():
            posterior = model_list_gp.posterior(train_x_raw)
        output_variables = [
            DistributionVariable(name="output1"),
            DistributionVariable(name="output2"),
            DistributionVariable(name="output3"),
            DistributionVariable(name="output4"),
        ]
        gp_lume_model = GPModel(
            model=model_list_gp,
            input_variables=input_variables,
            output_variables=output_variables,
        )
        output_dict = gp_lume_model.evaluate(input_dict)
        assert gp_lume_model.get_input_size() == 1
        assert gp_lume_model.get_output_size() == 4
        assert len(output_dict) == 4
        assert torch.allclose(output_dict["output1"].mean, posterior.mean[:, 0])
        assert torch.allclose(output_dict["output1"].variance, posterior.variance[:, 0])
        assert torch.allclose(output_dict["output2"].mean, posterior.mean[:, 1])
        assert torch.allclose(output_dict["output2"].variance, posterior.variance[:, 1])
        assert torch.allclose(output_dict["output3"].mean, posterior.mean[:, 2])
        assert torch.allclose(output_dict["output3"].variance, posterior.variance[:, 2])
        assert torch.allclose(output_dict["output4"].mean, posterior.mean[:, 3])
        assert torch.allclose(output_dict["output4"].variance, posterior.variance[:, 3])

    def test_transformers(
        self,
        create_single_task_gp,
        create_single_task_gp_w_transform,
        single_task_gp_model_kwargs,
    ):
        model, test_x = create_single_task_gp
        model_tf, _, input_transformer, output_transformer = (
            create_single_task_gp_w_transform
        )

        # Test with ModelListGP
        # Test with passing a list of transformers
        model_list_gp = ModelListGP(model, model)
        output_variables = [
            DistributionVariable(name="output1"),
            DistributionVariable(name="output2"),
        ]
        lume_model = GPModel(
            model=model_list_gp,
            input_variables=single_task_gp_model_kwargs["input_variables"],
            output_variables=output_variables,
            input_transformers=[input_transformer],
            output_transformers=[output_transformer],
        )
        # Predict with original model
        test_x_tf = input_transformer.transform(test_x)
        original_pred = model_list_gp.posterior(test_x_tf)
        original_mean = output_transformer.untransform(original_pred.mean)[0]
        original_variance = (
            abs(output_transformer.stdvs.squeeze(0)) ** 2 * original_pred.variance
        )
        # Predict with GPModel
        lume_pred = lume_model.evaluate({"input": test_x})
        assert lume_model.get_input_size() == 1
        assert lume_model.get_output_size() == 2
        assert len(lume_pred) == 2
        assert torch.allclose(original_mean[:, 0], lume_pred["output1"].mean)
        assert torch.allclose(original_variance[:, 0], lume_pred["output1"].variance)
        assert torch.allclose(original_mean[:, 1], lume_pred["output2"].mean)
        assert torch.allclose(original_variance[:, 1], lume_pred["output2"].variance)

        # Test with botorch transformers set as attributes
        # this should ignore any transformers passed to the GPModel
        model_list_gp = ModelListGP(model_tf, model_tf)
        output_variables = [
            DistributionVariable(name="output1"),
            DistributionVariable(name="output2"),
        ]
        lume_model = GPModel(
            model=model_list_gp,
            input_variables=single_task_gp_model_kwargs["input_variables"],
            output_variables=output_variables,
            input_transformers=[input_transformer],
            output_transformers=[output_transformer],
        )

        # Predict with original model
        # We have to transform twice to match the lume-model behavior in this case
        test_x_tf = input_transformer.transform(test_x)
        original_pred = model_list_gp.posterior(test_x_tf)
        original_mean = output_transformer.untransform(original_pred.mean)[0]
        original_variance = (
            abs(output_transformer.stdvs.squeeze(0)) ** 2 * original_pred.variance
        )
        # Predict with GPModel
        lume_pred = lume_model.evaluate({"input": test_x})
        assert lume_model.get_input_size() == 1
        assert lume_model.get_output_size() == 2
        assert len(lume_pred) == 2
        assert torch.allclose(original_mean[:, 0], lume_pred["output1"].mean)
        assert torch.allclose(original_variance[:, 0], lume_pred["output1"].variance)
        assert torch.allclose(original_mean[:, 1], lume_pred["output2"].mean)
        assert torch.allclose(original_variance[:, 1], lume_pred["output2"].variance)

        # Dump and load the model
        lume_model.dump("test.yml")
        loaded_model = GPModel("test.yml")
        lume_pred_loaded = loaded_model.evaluate({"input": test_x})
        assert torch.allclose(
            lume_pred["output1"].mean, lume_pred_loaded["output1"].mean
        )
        assert torch.allclose(
            lume_pred["output1"].variance, lume_pred_loaded["output1"].variance
        )
        assert torch.allclose(
            lume_pred["output2"].mean, lume_pred_loaded["output2"].mean
        )
        assert torch.allclose(
            lume_pred["output2"].variance, lume_pred_loaded["output2"].variance
        )
        os.remove("test.yml")
        os.remove("test_model.pt")
        os.remove("test_input_transformers_0.pt")
        os.remove("test_output_transformers_0.pt")



import pytest
from pydantic import ValidationError

from lume_model.variables import ScalarVariable, get_variable


class TestScalarVariable:
    def test_init(self):
        ScalarVariable(
            name="test",
            default_value=0.1,
            value_range=(0, 1),
            unit="m",
        )
        # missing name
        with pytest.raises(ValidationError):
            ScalarVariable(default_value=0.1, value_range=(0, 1))
        # constant variable
        ScalarVariable(
            name="test",
            default_value=0.5,
            value_range=None,
        )

    def test_validate_value(self):
        var = ScalarVariable(
            name="test",
            default_value=0.8,
            value_range=(0.0, 10.0),
            value_range_tolerance=1e-8,
        )
        var.validate_value(5.0)
        with pytest.raises(TypeError):
            var.validate_value(int(5))
        # test validation config
        var.validate_value(11.0, config="none")
        # range check with strictness flag
        with pytest.raises(ValueError):
            var.validate_value(11.0, config="error")
        # constant variable
        constant_var = ScalarVariable(
            name="test",
            default_value=1.3,
            is_constant=True,
            value_range_tolerance=1e-5,
        )
        constant_var.validate_value(1.3, config="error")
        with pytest.raises(ValueError):
            constant_var.validate_value(1.4, config="error")
        # test tolerance
        var.validate_value(10.0 + 1e-9, config="error")
        with pytest.raises(ValueError):
            var.validate_value(10.0 + 1e-7, config="error")
        constant_var.validate_value(1.3 + 1e-6, config="error")
        with pytest.raises(ValueError):
            constant_var.validate_value(1.3 + 1e-4, config="error")
        # test constant range validation
        with pytest.raises(ValueError):
            constant_var = ScalarVariable(
                name="test",
                default_value=1.3,
                is_constant=True,
                value_range_tolerance=1e-5,
                value_range=(1.3, 1.5),
            )


def test_get_variable():
    var = get_variable(ScalarVariable.__name__)(name="test")
    assert isinstance(var, ScalarVariable)


# @pytest.mark.parametrize(
#     "variable_name,default,value_range,axis_labels,x_min,y_min,x_max,y_max",
#     [
#         ("test", np.array([[1, 2,], [3, 4]]), [0, 1], ["x", "y"], 0, 0, 1, 1),
#         pytest.param(
#             "test", 1.0, [0, 1], ["x", "y"], 0, 0, 1, 1, marks=pytest.mark.xfail
#         ),
#         ("test", np.empty((3, 3)), [0, 1], ["x", "y"], 0, 0, 1, 1),
#     ],
# )
# def test_input_image_variable(
#     variable_name, default, value_range, axis_labels, x_min, y_min, x_max, y_max
# ):
#     # test correctly typed
#     ImageInputVariable(
#         name=variable_name,
#         default=default,
#         value_range=value_range,
#         axis_labels=axis_labels,
#         x_min=x_min,
#         y_min=y_min,
#         x_max=x_max,
#         y_max=y_max,
#     )
#
#     # test missing name
#     with pytest.raises(ValidationError):
#         ImageInputVariable(
#             default=default,
#             value_range=value_range,
#             axis_labels=axis_labels,
#             x_min=x_min,
#             y_min=y_min,
#             x_max=x_max,
#             y_max=y_max,
#         )
#
#     # test missing axis labels
#     with pytest.raises(ValidationError):
#         ImageInputVariable(
#             name=variable_name,
#             default=default,
#             value_range=value_range,
#             x_min=x_min,
#             y_min=y_min,
#             x_max=x_max,
#             y_max=y_max,
#         )


# @pytest.mark.parametrize(
#     "variable_name,default,axis_labels",
#     [
#         ("test", np.array([[1, 2,], [3, 4]]), ["x", "y"],),
#         pytest.param("test", 1.0, ["x", "y"], marks=pytest.mark.xfail),
#     ],
# )
# def test_output_image_variable(variable_name, default, axis_labels):
#     shape = default.shape
#     ImageOutputVariable(
#         name=variable_name, default=default, shape=shape, axis_labels=axis_labels,
#     )
#
#     # test missing name
#     with pytest.raises(ValidationError):
#         ImageOutputVariable(
#             default=default, shape=shape, axis_labels=axis_labels,
#         )
#
#     # test missing axis labels
#     with pytest.raises(ValidationError):
#         ImageOutputVariable(
#             name=variable_name, default=default,
#         )
#
#     # test missing value
#     ImageOutputVariable(
#         name=variable_name, axis_labels=axis_labels,
#     )


# @pytest.mark.parametrize(
#     "variable_name,default,value_range,axis_labels,x_min,y_min,x_max,y_max",
#     [
#         ("test", np.array([[1, 2,], [3, 4]]), [0, 1], ["x", "y"], 0, 0, 1, 1),
#         pytest.param(
#             "test", 1.0, [0, 1], ["x", "y"], 0, 0, 1, 1, marks=pytest.mark.xfail
#         ),
#     ],
# )
# def test_image_variable_shape(
#     variable_name, default, value_range, axis_labels, x_min, y_min, x_max, y_max
# ):
#     shape = default.shape
#
#     # test correctly typed
#     variable = ImageInputVariable(
#         name=variable_name,
#         default=default,
#         value_range=value_range,
#         axis_labels=axis_labels,
#         x_min=x_min,
#         y_min=y_min,
#         x_max=x_max,
#         y_max=y_max,
#     )
#
#     assert shape == variable.shape


# @pytest.mark.parametrize(
#     "variable_name,default,value_range,axis_labels,x_min,y_min,x_max,y_max",
#     [("test", np.array([[1, 2,], [3, 4]]), [0, 1], ["x", "y"], 0, 0, 1, 1)],
# )
# def test_input_image_variable_color_mode(
#     variable_name, default, value_range, axis_labels, x_min, y_min, x_max, y_max
# ):
#
#     random_rgb_default = np.random.rand(10, 10, 3)
#
#     # test correctly typed
#     variable = ImageInputVariable(
#         name=variable_name,
#         default=random_rgb_default,
#         value_range=value_range,
#         axis_labels=axis_labels,
#         x_min=x_min,
#         y_min=y_min,
#         x_max=x_max,
#         y_max=y_max,
#     )
#
#     with pytest.raises(ValueError):
#         random_rgb_default = np.random.rand(10, 10, 2)
#         # test correctly typed
#         variable = ImageInputVariable(
#             name=variable_name,
#             default=random_rgb_default,
#             value_range=value_range,
#             axis_labels=axis_labels,
#             x_min=x_min,
#             y_min=y_min,
#             x_max=x_max,
#             y_max=y_max,
#         )


# @pytest.mark.parametrize(
#     "variable_name,default,value_range,dim_labels",
#     [
#         ("test", np.array([[1, 2,], [3, 4]]), [0, 5], ["x, y"]),
#         pytest.param("test", [0, 1], [0, 5], ["x", "y"], marks=pytest.mark.xfail),
#     ],
# )
# def test_input_array_variable(variable_name, default, value_range, dim_labels):
#     # test correctly typed
#     ArrayInputVariable(
#         name=variable_name,
#         default=default,
#         value_range=value_range,
#         dim_labels=dim_labels,
#     )
#
#     # test missing name
#     with pytest.raises(ValidationError):
#         ArrayInputVariable(
#             default=default, value_range=value_range, dim_labels=dim_labels,
#         )
#
#     # test missing axis labels
#     ArrayInputVariable(
#         name=variable_name, default=default, value_range=value_range,
#     )


# @pytest.mark.parametrize(
#     "variable_name,default,dim_labels",
#     [
#         ("test", np.array([[1, 2,], [3, 4]]), ["x", "y"],),
#         pytest.param("test", 1.0, ["x", "y"], marks=pytest.mark.xfail),
#     ],
# )
# def test_output_array_variable(variable_name, default, dim_labels):
#     shape = default.shape
#     ArrayOutputVariable(
#         name=variable_name, default=default, shape=shape, dim_labels=dim_labels,
#     )
#
#     # test missing name
#     with pytest.raises(ValidationError):
#         ArrayOutputVariable(
#             default=default, shape=shape, dim_labels=dim_labels,
#         )
#
#     # test missing labels
#     ArrayOutputVariable(
#         name=variable_name, default=default,
#     )
#
#     # test missing value
#     ArrayOutputVariable(
#         name=variable_name, dim_labels=dim_labels,
#     )


# @pytest.mark.parametrize(
#     "rows,variables",
#     [
#         (
#             None,
#             {
#                 "col1": {
#                     "row1": ScalarInputVariable(
#                         name="col1_row1", default=0,value_range=[-1, -1]
#                     ),
#                     "row2": ScalarInputVariable(
#                         name="col1_row2", default=0,value_range=[-1, 1]
#                     ),
#                 },
#                 "col2": {
#                     "row1": ScalarInputVariable(
#                         name="col2_row1", default=0,value_range=[-1, -1]
#                     ),
#                     "row2": ScalarInputVariable(
#                         name="col2_row2", default=0,value_range=[-1, 1]
#                     ),
#                 },
#             },
#         ),
#         pytest.param(
#             ["row1", "row2"],
#             {
#                 "col1": {
#                     "row1": ScalarInputVariable(
#                         name="col1_row1", default=0, value_range=[-1, -1]
#                     ),
#                     "row2": 5,
#                 },
#                 "col2": {
#                     "row1": ScalarInputVariable(
#                         name="col2_row1", default=0,value_range=[-1, -1]
#                     ),
#                     "row2": ScalarInputVariable(
#                         name="col2_row2", default=0,value_range=[-1, 1]
#                     ),
#                 },
#             },
#             marks=pytest.mark.xfail,
#         ),
#         pytest.param(
#             None,
#             {
#                 "col1": ArrayInputVariable(
#                     name="test", default=np.array([1, 2]), value_range=[0, 10]
#                 ),
#                 "col2": {
#                     "row1": ScalarInputVariable(
#                         name="col2_row1", default=0, value_range=[-1, -1]
#                     ),
#                     "row2": ScalarInputVariable(
#                         name="col2_row2", default=0, value_range=[-1, 1]
#                     ),
#                 },
#             },
#             marks=pytest.mark.xfail,
#         ),
#         (
#             ["row1", "row2"],
#             {
#                 "col1": ArrayInputVariable(
#                     name="test", default=np.array([1, 2]), value_range=[0, 10]
#                 ),
#                 "col2": {
#                     "row1": ScalarInputVariable(
#                         name="col2_row1", default=0, value_range=[-1, -1]
#                     ),
#                     "row2": ScalarInputVariable(
#                         name="col2_row2", default=0, value_range=[-1, 1]
#                     ),
#                 },
#             },
#         ),
#         pytest.param(
#             ["row1", "row2"],
#             {
#                 "col1": ArrayInputVariable(
#                     name="test", default=np.array([1, 2, 3, 4]), value_range=[0, 10]
#                 ),
#                 "col2": {
#                     "row1": ScalarInputVariable(
#                         name="col2_row1", default=0, value_range=[-1, -1]
#                     ),
#                     "row2": ScalarInputVariable(
#                         name="col2_row2", default=0, value_range=[-1, 1]
#                     ),
#                 },
#             },
#             marks=pytest.mark.xfail,
#         ),
#     ],
# )
# def test_variable_table(rows, variables):
#     if rows:
#         table_var = TableVariable(table_rows=rows, table_data=variables)
#     else:
#         table_var = TableVariable(table_data=variables)
#
#     with pytest.raises(ValueError):
#         table_var = TableVariable(table_data=None)



import os
import json
from typing import Any, Union

import pytest

from lume_model.utils import variables_from_yaml
from lume_model.variables import ScalarVariable, DistributionVariable

try:
    import torch
    from botorch.models.transforms.input import AffineInputTransform  # noqa: F401
    from botorch.models.transforms.outcome import Standardize  # noqa: F401
    from botorch.models import MultiTaskGP, SingleTaskGP
    from lume_model.models import TorchModel, TorchModule
except ModuleNotFoundError:
    pass


@pytest.fixture(scope="session")
def rootdir() -> str:
    return os.path.dirname(os.path.abspath(__file__))


# TorchModel fixtures
@pytest.fixture(scope="session")
def simple_variables() -> dict[str, Union[list[ScalarVariable], list[ScalarVariable]]]:
    input_variables = [
        ScalarVariable(name="input1", default_value=1.0, value_range=(0.0, 5.0)),
        ScalarVariable(name="input2", default_value=2.0, value_range=(1.0, 3.0)),
    ]
    output_variables = [ScalarVariable(name="output1"), ScalarVariable(name="output2")]
    return {"input_variables": input_variables, "output_variables": output_variables}


@pytest.fixture(scope="module")
def california_model_info(rootdir) -> dict[str, str]:
    try:
        with open(
            f"{rootdir}/test_files/california_regression/model_info.json", "r"
        ) as f:
            model_info = json.load(f)
        return model_info
    except FileNotFoundError as e:
        pytest.skip(str(e))


@pytest.fixture(scope="module")
def california_variables(rootdir) -> tuple[list[ScalarVariable], list[ScalarVariable]]:
    try:
        file = f"{rootdir}/test_files/california_regression/variables.yml"
        input_variables, output_variables = variables_from_yaml(file)
        return input_variables, output_variables
    except FileNotFoundError as e:
        pytest.skip(str(e))


@pytest.fixture(scope="module")
def california_transformers(rootdir):
    botorch = pytest.importorskip("botorch")

    try:
        with open(
            f"{rootdir}/test_files/california_regression/normalization.json", "r"
        ) as f:
            normalizations = json.load(f)
    except FileNotFoundError as e:
        pytest.skip(str(e))

    input_transformer = botorch.models.transforms.input.AffineInputTransform(
        len(normalizations["x_mean"]),
        coefficient=torch.tensor(normalizations["x_scale"]),
        offset=torch.tensor(normalizations["x_mean"]),
    )
    output_transformer = botorch.models.transforms.input.AffineInputTransform(
        len(normalizations["y_mean"]),
        coefficient=torch.tensor(normalizations["y_scale"]),
        offset=torch.tensor(normalizations["y_mean"]),
    )
    return input_transformer, output_transformer


@pytest.fixture(scope="module")
def california_model_kwargs(
    rootdir,
    california_model_info,
    california_variables,
    california_transformers,
) -> dict[str, Any]:
    _ = pytest.importorskip("botorch")

    input_variables, output_variables = california_variables
    input_transformer, output_transformer = california_transformers
    model_kwargs = {
        "model": torch.load(
            f"{rootdir}/test_files/california_regression/model.pt", weights_only=False
        ),
        "input_variables": input_variables,
        "output_variables": output_variables,
        "input_transformers": [input_transformer],
        "output_transformers": [output_transformer],
        "output_format": "tensor",
    }
    return model_kwargs


@pytest.fixture(scope="module")
def california_test_input_tensor(rootdir: str):
    torch = pytest.importorskip("torch")

    try:
        test_input_tensor = torch.load(
            f"{rootdir}/test_files/california_regression/test_input_tensor.pt",
            weights_only=False,
        )
    except FileNotFoundError as e:
        pytest.skip(str(e))
    return test_input_tensor


@pytest.fixture(scope="module")
def california_test_input_dict(
    california_test_input_tensor, california_model_info
) -> dict:
    pytest.importorskip("botorch")

    test_input_dict = {
        key: california_test_input_tensor[0, idx]
        for idx, key in enumerate(california_model_info["model_in_list"])
    }
    return test_input_dict


@pytest.fixture(scope="module")
def california_model(california_model_kwargs):
    _ = pytest.importorskip("botorch")

    return TorchModel(**california_model_kwargs)


@pytest.fixture(scope="module")
def california_module(california_model):
    _ = pytest.importorskip("botorch")

    return TorchModule(model=california_model)


# GPModel fixtures
@pytest.fixture(scope="session")
def gp_variables() -> dict[
    str, Union[list[ScalarVariable], list[DistributionVariable]]
]:
    input_variables = [ScalarVariable(name="input")]
    output_variables = [
        DistributionVariable(name="output1"),
        DistributionVariable(name="output2"),
    ]
    return input_variables, output_variables


# SingleTask GP
@pytest.fixture(scope="module")
def single_task_gp_transformers(rootdir):
    input_transformer = torch.load(
        f"{rootdir}/test_files/single_task_gp/input_transformers.pt", weights_only=False
    )
    output_transformer = torch.load(
        f"{rootdir}/test_files/single_task_gp/output_transformers.pt",
        weights_only=False,
    )
    return input_transformer, output_transformer


@pytest.fixture(scope="module")
def single_task_gp_model_kwargs(
    rootdir,
    gp_variables,
    single_task_gp_transformers,
) -> dict[str, Any]:
    _ = pytest.importorskip("botorch")

    input_variables, output_variables = gp_variables
    input_transformer, output_transformer = single_task_gp_transformers
    model_kwargs = {
        "model": torch.load(
            f"{rootdir}/test_files/single_task_gp/model.pt", weights_only=False
        ),
        "input_variables": input_variables,
        "output_variables": output_variables,
        "input_transformers": [input_transformer],
        "output_transformers": [output_transformer],
    }
    return model_kwargs


# MultiTask GP
@pytest.fixture(scope="module")
def multi_task_gp_transformers(rootdir):
    input_transformer = torch.load(
        f"{rootdir}/test_files/multi_task_gp/input_transformers.pt", weights_only=False
    )
    output_transformer = torch.load(
        f"{rootdir}/test_files/multi_task_gp/output_transformers.pt", weights_only=False
    )
    return input_transformer, output_transformer


@pytest.fixture(scope="module")
def multi_task_gp_model_kwargs(
    rootdir,
    gp_variables,
    multi_task_gp_transformers,
) -> dict[str, Any]:
    _ = pytest.importorskip("botorch")

    input_variables, output_variables = gp_variables
    input_transformer, output_transformer = multi_task_gp_transformers
    model_kwargs = {
        "model": torch.load(
            f"{rootdir}/test_files/multi_task_gp/model.pt", weights_only=False
        ),
        "input_variables": input_variables,
        "output_variables": output_variables,
        "input_transformers": [input_transformer],
        "output_transformers": [output_transformer],
    }
    return model_kwargs


# ModelListGP
@pytest.fixture(scope="module")
def create_multi_task_gp():
    _ = pytest.importorskip("botorch")
    tkwargs = {"dtype": torch.double}
    train_x_raw, train_y = get_random_data(
        batch_shape=torch.Size(), m=1, n=10, **tkwargs
    )
    task_idx = torch.cat(
        [torch.ones(5, 1, **tkwargs), torch.zeros(5, 1, **tkwargs)], dim=0
    )
    train_x = torch.cat([train_x_raw, task_idx], dim=-1)
    # single output
    model = MultiTaskGP(
        train_X=train_x,
        train_Y=train_y,
        task_feature=-1,
        output_tasks=[0],
    )
    # multi output
    model2 = MultiTaskGP(
        train_X=train_x,
        train_Y=train_y,
        task_feature=-1,
    )
    return model, model2, train_x_raw


@pytest.fixture(scope="module")
def create_single_task_gp():
    tkwargs = {"dtype": torch.double}
    train_x1, train_y1 = get_random_data(batch_shape=torch.Size(), m=1, n=10, **tkwargs)
    model1 = SingleTaskGP(train_X=train_x1, train_Y=train_y1, outcome_transform=None)
    model1.to(**tkwargs)
    test_x = torch.tensor([[0.25], [0.75]], **tkwargs)
    return model1, test_x


@pytest.fixture(scope="module")
def create_single_task_gp_w_transform():
    tkwargs = {"dtype": torch.double}
    train_x1, train_y1 = get_random_data(batch_shape=torch.Size(), m=1, n=10, **tkwargs)
    input_transform = AffineInputTransform(
        1,
        coefficient=train_x1.std(dim=0),
        offset=train_y1.mean(dim=0),
    )
    output_transform = Standardize(m=1)
    model1 = SingleTaskGP(
        train_X=train_x1,
        train_Y=train_y1,
        input_transform=input_transform,
        outcome_transform=output_transform,
    )
    model1.to(**tkwargs)
    test_x = torch.tensor([[0.25], [0.75]], **tkwargs)
    return model1, test_x, input_transform, output_transform


def get_random_data(
    batch_shape: torch.Size, m: int, d: int = 1, n: int = 10, **tkwargs
):
    r"""Generate random data for testing purposes.

    Args:
        batch_shape: The batch shape of the data.
        m: The number of outputs.
        d: The dimension of the input.
        n: The number of data points.
        tkwargs: `device` and `dtype` tensor constructor kwargs.

    Returns:
        A tuple `(train_X, train_Y)` with randomly generated training data.
    """
    rep_shape = batch_shape + torch.Size([1, 1])
    train_x = torch.stack(
        [torch.linspace(0, 0.95, n, **tkwargs) for _ in range(d)], dim=-1
    )
    train_x = train_x + 0.05 * torch.rand_like(train_x).repeat(rep_shape)
    train_x[0] += 0.02  # modify the first batch
    train_y = torch.sin(train_x[..., :1] * (2 * torch.pi))
    train_y = train_y + 0.2 * torch.randn(n, m, **tkwargs).repeat(rep_shape)
    return train_x, train_y



import os
import random
import warnings
from copy import deepcopy

import pytest

try:
    import torch
    from botorch.models import SingleTaskGP
    from lume_model.models import TorchModel, TorchModule

    torch.manual_seed(42)
except ImportError:
    pass

random.seed(42)


def assert_california_module_result(result: torch.Tensor, idx=None):
    target = torch.tensor(
        [4.0636503726, 2.7774916915, 2.7928111793], dtype=result.dtype
    )
    if idx is None:
        assert all(torch.isclose(result, target))
    else:
        assert torch.isclose(result, target[idx])


def assert_model_equality(m1: TorchModel, m2: TorchModel):
    assert m1.input_variables == m2.input_variables
    assert m1.output_variables == m2.output_variables
    assert str(m1.model.state_dict()) == str(m2.model.state_dict())
    for attr in ["input_transformers", "output_transformers"]:
        m1_transformers, m2_transformers = getattr(m1, attr), getattr(m2, attr)
        assert len(m1_transformers) == len(m2_transformers)
        for i, t in enumerate(m1_transformers):
            if isinstance(t, torch.nn.Module):
                assert isinstance(m2_transformers[i], torch.nn.Module)
                assert str(t.state_dict()) == str(m2_transformers[i].state_dict())
    assert m1.output_format == m2.output_format
    assert m1.device == m2.device
    assert m1.fixed_model == m2.fixed_model


def assert_module_equality(m1: TorchModule, m2: TorchModule):
    assert_model_equality(m1.model, m2.model)
    assert m1.input_order == m2.input_order
    assert m1.output_order == m2.output_order


class TestTorchModule:
    def test_module_initialization(self, california_model):
        lume_module = TorchModule(model=california_model)
        parameters_with_requires_grad = [
            param for param in lume_module.parameters() if param.requires_grad
        ]

        # gradients should be deactivated and module in eval mode
        assert not parameters_with_requires_grad
        assert not lume_module.training

    def test_module_parameters_match_model(self, california_model):
        california_module = TorchModule(model=california_model)
        params_match = [
            a == b
            for a, b in zip(
                california_module.parameters(), california_model.model.parameters()
            )
        ]

        assert len(list(california_module.parameters())) == len(
            list(california_model.model.parameters())
        )
        for param_match in params_match:
            if isinstance(param_match, torch.Tensor):
                assert torch.all(param_match)
            else:
                assert param_match

    def test_module_train_and_eval_mode(self, california_module):
        assert california_module.training is False
        assert california_module._model.model.training is False
        california_module.train()
        assert california_module.training is True
        assert california_module._model.model.training is True
        california_module.eval()
        assert california_module.training is False
        assert california_module._model.model.training is False

    def test_module_differentiability(
        self, california_test_input_tensor, california_module
    ):
        lume_module = deepcopy(california_module)
        lume_module.train()
        lume_module.requires_grad_(True)
        parameters_with_requires_grad = [
            param for param in lume_module.parameters() if param.requires_grad
        ]
        criterion = torch.nn.MSELoss()

        assert lume_module.training
        for param in lume_module.parameters():
            assert param.requires_grad
        for param in lume_module._model.model.parameters():
            assert param.requires_grad
        assert len(parameters_with_requires_grad) == 8
        outputs = lume_module(california_test_input_tensor)
        loss = criterion(outputs, torch.zeros(outputs.shape, dtype=outputs.dtype))
        loss.backward()

    def test_module_from_yaml(self, rootdir: str, california_module):
        file = os.path.join(
            rootdir, "test_files", "california_regression", "torch_module.yml"
        )
        yaml_module = TorchModule(file)

        assert_module_equality(yaml_module, california_module)

    def test_module_as_yaml(self, rootdir: str, california_module):
        filename = "test_torch_module"
        file = f"{filename}.yml"
        california_module.dump(file)
        yaml_module = TorchModule(file)
        assert_module_equality(yaml_module, california_module)
        os.remove(file)
        os.remove(f"{filename}_model.pt")
        os.remove(f"{filename}_input_transformers_0.pt")
        os.remove(f"{filename}_output_transformers_0.pt")

    def test_module_as_jit(self, rootdir: str, california_module):
        filename = "test_torch_module"
        file = f"{filename}.yml"

        # Test saving with save_jit=True
        california_module.dump(file, save_jit=True)
        assert os.path.exists(f"{filename}_model.jit")

        # Test loading the JIT model
        try:
            loaded_model = torch.jit.load(f"{filename}_model.jit")
            assert loaded_model is not None
        except Exception as e:
            pytest.fail(f"Failed to load JIT model: {e}")

        os.remove(file)
        os.remove(f"{filename}_model.pt")
        os.remove(f"{filename}_model.jit")
        os.remove(f"{filename}_input_transformers_0.pt")
        os.remove(f"{filename}_output_transformers_0.pt")

    def test_module_call_single_input(
        self, california_test_input_tensor, california_model
    ):
        lume_module = TorchModule(
            model=california_model,
            input_order=[california_model.input_names[0]],
        )
        input_tensor = deepcopy(
            california_test_input_tensor[:, 0].unsqueeze(-1)
        )  # shape (3, 1)
        result = lume_module(input_tensor)
        target = torch.tensor(
            [3.5094612847, 1.7297480438, 2.7042855903], dtype=result.dtype
        )

        assert tuple(result.size()) == (3,)
        assert all(torch.isclose(result, target))

    def test_module_call_single_input_bad_shape(
        self, california_test_input_tensor, california_model
    ):
        lume_module = TorchModule(
            model=california_model,
            input_order=[california_model.input_names[0]],
        )
        input_tensor = deepcopy(california_test_input_tensor[:, 0])  # shape (3,)

        with pytest.raises(ValueError):
            lume_module(input_tensor)

    def test_module_call_single_sample(
        self, california_test_input_tensor, california_module
    ):
        idx = 0
        input_tensor = deepcopy(california_test_input_tensor[idx, :]).unsqueeze(
            0
        )  # shape (1,8)
        result = california_module(input_tensor)

        assert tuple(result.size()) == ()
        assert_california_module_result(result, idx=idx)

    def test_module_call_n_samples(
        self, california_test_input_tensor, california_module
    ):
        result = california_module(california_test_input_tensor)

        assert tuple(result.size()) == (3,)
        assert_california_module_result(result)

    def test_module_reordered_inputs(
        self, california_test_input_tensor, california_model
    ):
        # shuffle the input names and the values associated with them
        idx = 0
        shuffled_inputs = list(
            zip(california_model.input_names, california_test_input_tensor[idx])
        )
        random.shuffle(shuffled_inputs)
        input_names = [shuffled_inputs[i][0] for i in range(len(shuffled_inputs))]
        input_tensor = torch.tensor(
            [shuffled_inputs[i][1] for i in range(len(shuffled_inputs))]
        ).reshape(1, -1)
        output_names = deepcopy(california_model.output_names)
        lume_module = TorchModule(
            model=california_model,
            input_order=input_names,
            output_order=output_names,
        )
        result = lume_module(input_tensor)

        assert tuple(result.size()) == ()
        assert_california_module_result(result, idx=idx)

    def test_module_call_manipulate_output(
        self, california_test_input_tensor, california_model
    ):
        n = 2
        output_order = deepcopy(california_model.output_names)
        output_order.append(f"MedHouseVal_x{n}")

        class ExampleTorchModule(TorchModule):
            def __init__(self, factor: int, **kwargs):
                super().__init__(**kwargs)
                self.factor = factor

            def manipulate_output(self, y_model: dict[str, torch.Tensor]):
                y_model[self.output_order[-1]] = y_model["MedHouseVal"] * self.factor
                return y_model

        lume_module = ExampleTorchModule(
            factor=n,
            model=california_model,
            input_order=california_model.input_names,
            output_order=output_order,
        )
        input_tensor = deepcopy(california_test_input_tensor)
        result = lume_module(input_tensor)

        assert tuple(result.size()) == (3, 2)
        assert_california_module_result(result[:, 0])
        assert_california_module_result(result[:, 1] / n)

    def test_module_call_batch_n_samples(
        self, california_test_input_tensor, california_module
    ):
        # module should be able to handle input of shape [n_batch, n_samples, n_dim]
        n_batch = 5
        input_tensor = california_test_input_tensor.unsqueeze(0).repeat((n_batch, 1, 1))
        result = california_module(input_tensor)

        assert tuple(result.shape) == (n_batch, 3)
        for i in range(n_batch):
            assert_california_module_result(result[i])

    def test_module_as_gp_prior_mean(
        self, california_test_input_tensor, california_module
    ):
        train_x = california_test_input_tensor.double()
        train_y = california_module(train_x).unsqueeze(-1)
        with warnings.catch_warnings():
            warnings.simplefilter(
                "ignore"
            )  # ignore warning that input data is not standardized
            gp = SingleTaskGP(
                train_x, train_y, mean_module=california_module, outcome_transform=None
            )
        x_lim = torch.stack(
            [torch.min(train_x, dim=0).values, torch.max(train_x, dim=0).values]
        )
        x_i = [
            x_lim[0, i] + (x_lim[1, i] - x_lim[0, i]) * torch.rand(size=(2,))
            for i in range(x_lim.shape[-1])
        ]
        test_x = torch.cartesian_prod(*x_i)
        test_y = california_module(test_x)
        post = gp.posterior(test_x)
        mean = post.mean.squeeze()

        assert tuple(mean.shape) == (test_x.shape[0],)
        assert tuple(test_y.shape) == (test_x.shape[0],)
        # GP should just predict the prior mean values for the posterior
        assert all(torch.isclose(mean, test_y))



import os

import pytest

from lume_model.utils import (
    verify_unique_variable_names,
    variables_as_yaml,
    variables_from_yaml,
)


def test_verify_unique_variable_names(simple_variables):
    input_variables = simple_variables["input_variables"]
    output_variables = simple_variables["output_variables"]
    # unique variables names
    verify_unique_variable_names(input_variables)
    verify_unique_variable_names(output_variables)
    # non-unique input names
    original_name = input_variables[1].name
    input_variables[1].name = input_variables[0].name
    with pytest.raises(ValueError):
        verify_unique_variable_names(input_variables)
    input_variables[1].name = original_name
    # non-unique output names
    original_name = output_variables[1].name
    output_variables[1].name = output_variables[0].name
    with pytest.raises(ValueError):
        verify_unique_variable_names(output_variables)
    output_variables[1].name = original_name


def test_variables_as_yaml(simple_variables):
    file = "test_variables.yml"
    variables_as_yaml(**simple_variables, file=file)
    os.remove(file)


def test_variables_as_and_from_yaml(simple_variables):
    file = "test_variables.yml"
    variables_as_yaml(**simple_variables, file=file)
    variables = variables_from_yaml(file)
    os.remove(file)
    assert simple_variables["input_variables"] == variables[0]
    assert simple_variables["output_variables"] == variables[1]



import os
import io
import sys
import pytest
import yaml

from lume_model.base import LUMEBaseModel
from lume_model.variables import ScalarVariable


class ExampleModel(LUMEBaseModel):
    def _evaluate(self, input_dict):
        pass


class TestBaseModel:
    def test_init(self, simple_variables):
        # init with no variable specification
        with pytest.raises(TypeError):
            _ = LUMEBaseModel()

        # init child class with no _evaluate function
        class NoEvaluateModel(LUMEBaseModel):
            def predict(self, input_dict):
                pass

        with pytest.raises(TypeError):
            _ = NoEvaluateModel(**simple_variables)

        # TODO: move to test_torch_model.py
        # # init child class with input variables missing default value
        # simple_variables_no_default = copy.deepcopy(simple_variables)
        # simple_variables_no_default["input_variables"][0].default_value = None
        # with pytest.raises(ValueError):
        #     _ = ExampleModel(**simple_variables_no_default)

        # init child class with evaluate function
        example_model = ExampleModel(**simple_variables)
        assert example_model.input_variables == simple_variables["input_variables"]
        assert example_model.output_variables == simple_variables["output_variables"]

        # input and output variables sharing names is fine
        input_variables = simple_variables["input_variables"]
        output_variables = simple_variables["output_variables"]
        original_name = input_variables[0].name
        input_variables[0].name = output_variables[0].name
        _ = ExampleModel(**simple_variables)
        input_variables[0].name = original_name

    def test_dict(self, simple_variables):
        example_model = ExampleModel(**simple_variables)
        dict_output = example_model.model_dump()
        assert isinstance(dict_output["input_variables"], list)
        assert isinstance(dict_output["output_variables"], list)
        assert len(dict_output["input_variables"]) == 2

    def test_json(self, simple_variables):
        example_model = ExampleModel(**simple_variables)
        _ = example_model.json()

    def test_yaml_serialization(self, simple_variables):
        example_model = ExampleModel(**simple_variables)
        yaml_output = example_model.yaml()
        dict_output = yaml.safe_load(yaml_output)
        dict_output["input_variables"]["input1"]["variable_class"] = (
            ScalarVariable.__name__
        )

        # test loading from yaml
        loaded_model = ExampleModel(**dict_output)
        assert loaded_model == example_model

    def test_file_serialization(self, simple_variables):
        example_model = ExampleModel(**simple_variables)
        file = "test_model.yml"
        example_model.dump(file)

        os.remove(file)

    def test_deserialization_from_config(self, simple_variables):
        example_model = ExampleModel(**simple_variables)
        file = "test_model.yml"
        _ = example_model.dump(file)
        loaded_model = ExampleModel(file)
        os.remove(file)
        assert loaded_model.input_variables == example_model.input_variables
        assert loaded_model.output_variables == example_model.output_variables

    def test_input_names(self, simple_variables):
        example_model = ExampleModel(**simple_variables)
        for i, var in enumerate(simple_variables["input_variables"]):
            assert example_model.input_names.index(var.name) == i

    def test_output_names(self, simple_variables):
        example_model = ExampleModel(**simple_variables)
        for i, var in enumerate(simple_variables["output_variables"]):
            assert example_model.output_names.index(var.name) == i

    def test_input_validation(self, simple_variables, monkeypatch):
        example_model = ExampleModel(**simple_variables)
        input_variables = simple_variables["input_variables"]
        input_dict = {input_variables[0].name: 2.0, input_variables[1].name: 1.5}
        example_model.input_validation(input_dict)
        with pytest.raises(TypeError):
            input_dict[input_variables[0].name] = True
            example_model.input_validation(input_dict)

        # setting strictness flag
        assert input_variables[0].default_validation_config == "warn"
        with pytest.raises(ValueError):
            # has to be a ConfigEnum type
            example_model.input_validation_config = {input_variables[0].name: "test"}

        # range check with strictness flag
        example_model.input_validation_config = {input_variables[0].name: "error"}
        with pytest.raises(ValueError):
            input_dict[input_variables[0].name] = 6.0
            example_model.input_validation(input_dict)

        # a warning is printed
        example_model.input_validation_config = {input_variables[0].name: "warn"}
        input_dict[input_variables[0].name] = 6.0
        fake_out = io.StringIO()
        monkeypatch.setattr(sys, "stdout", fake_out)
        example_model.input_validation(input_dict)
        assert "Warning" in fake_out.getvalue()

        # nothing is printed/raised
        example_model.input_validation_config = {input_variables[0].name: "none"}
        input_dict[input_variables[0].name] = 6.0
        example_model.input_validation(input_dict)

    def test_output_validation(self, simple_variables):
        example_model = ExampleModel(**simple_variables)
        output_variables = simple_variables["output_variables"]
        output_dict = {output_variables[0].name: 3.0, output_variables[1].name: 1.7}
        example_model.output_validation(output_dict)
        with pytest.raises(TypeError):
            output_dict[output_variables[0].name] = "test"
            example_model.output_validation(output_dict)



"""
This module contains definitions of LUME-model variables for use with lume tools.
Variables are designed as pure descriptors and thus aren't intended to hold actual values,
but they can be used to validate encountered values.

For now, only scalar floating-point variables are supported.
"""

from abc import ABC, abstractmethod
from typing import Any, Optional, Type
from enum import Enum
import math

import numpy as np
from torch.distributions import Distribution as TDistribution
from pydantic import BaseModel, field_validator, model_validator, ConfigDict


class ConfigEnum(str, Enum):
    """Enum for configuration options during validation."""

    NULL = "none"
    WARN = "warn"
    ERROR = "error"


class Variable(BaseModel, ABC):
    """Abstract variable base class.

    Attributes:
        name: Name of the variable.
    """

    name: str

    @property
    @abstractmethod
    def default_validation_config(self) -> ConfigEnum:
        """Determines default behavior during validation."""
        return None

    @abstractmethod
    def validate_value(self, value: Any, config: dict[str, bool] = None):
        pass

    def model_dump(self, **kwargs) -> dict[str, Any]:
        config = super().model_dump(**kwargs)
        return {"variable_class": self.__class__.__name__} | config


class ScalarVariable(Variable):
    """Variable for float values.

    Attributes:
        default_value: Default value for the variable. Note that the LUMEBaseModel requires this
          for input variables, but it is optional for output variables.
        value_range: Value range that is considered valid for the variable. If the value range is set to None,
          the variable is interpreted as a constant and values are validated against the default value.
        is_constant: Flag indicating whether the variable is constant.
        unit: Unit associated with the variable.
    """

    model_config = ConfigDict(arbitrary_types_allowed=True)

    default_value: Optional[float] = None
    is_constant: Optional[bool] = False
    value_range: Optional[tuple[float, float]] = None
    # tolerance for floating point errors, currently only used for constant variables
    value_range_tolerance: Optional[float] = 1e-8
    unit: Optional[str] = None

    @field_validator("value_range", mode="before")
    @classmethod
    def validate_value_range(cls, value):
        if value is not None:
            value = tuple(value)
            if not value[0] <= value[1]:
                raise ValueError(
                    f"Minimum value ({value[0]}) must be lower or equal than maximum ({value[1]})."
                )
        return value

    @model_validator(mode="after")
    def validate_default_value(self):
        if self.default_value is not None and self.value_range is not None:
            if not self._value_is_within_range(self.default_value):
                raise ValueError(
                    "Default value ({}) is out of valid range: ([{},{}]).".format(
                        self.default_value, *self.value_range
                    )
                )
        return self

    @model_validator(mode="after")
    def validate_constant_value_range(self):
        if self.is_constant and self.value_range is not None:
            # if the upper limit is not equal to the lower limit, raise an error
            if not self.value_range[0] == self.value_range[1]:
                error_message = (
                    f"Expected range to be constant for constant variable '{self.name}', "
                    f"but received a range of values. Set range to None or set the "
                    f"upper limit equal to the lower limit."
                )
                raise ValueError(error_message)
        return self

    @property
    def default_validation_config(self) -> ConfigEnum:
        return "warn"

    def validate_value(self, value: float, config: ConfigEnum = None):
        """
        Validates the given value.

        Args:
            value (float): The value to be validated.
            config (ConfigEnum, optional): The configuration for validation. Defaults to None.
              Allowed values are "none", "warn", and "error".

        Raises:
            TypeError: If the value is not of type float.
            ValueError: If the value is out of the valid range or does not match the default value
              for constant variables.
        """
        _config = self.default_validation_config if config is None else config
        # mandatory validation
        self._validate_value_type(value)
        # optional validation
        if config != "none":
            self._validate_value_is_within_range(value, config=_config)

    @staticmethod
    def _validate_value_type(value: float):
        if not isinstance(value, float):
            raise TypeError(
                f"Expected value to be of type {float} or {np.float64}, "
                f"but received {type(value)}."
            )

    def _validate_value_is_within_range(self, value: float, config: ConfigEnum = None):
        if not self._value_is_within_range(value):
            if self.is_constant:
                error_message = "Expected value to be ({}) for constant variable ({}), but received ({}).".format(
                    self.default_value, self.name, value
                )
            else:
                error_message = (
                    "Value ({}) of '{}' is out of valid range: ([{},{}]).".format(
                        value, self.name, *self.value_range
                    )
                )
            range_warning_message = (
                error_message
                + " Executing the model outside of the training data range may result in"
                " unpredictable and invalid predictions."
            )
            if config == "warn":
                print("Warning: " + range_warning_message)
            else:
                raise ValueError(error_message)

    def _value_is_within_range(self, value) -> bool:
        self.value_range = self.value_range or (-np.inf, np.inf)
        tolerances = {"rel_tol": 0, "abs_tol": self.value_range_tolerance}
        is_within_range, is_within_tolerance = False, False
        # constant variables
        if self.value_range is None or self.is_constant:
            if self.default_value is None:
                is_within_tolerance = True
            else:
                is_within_tolerance = math.isclose(
                    value, self.default_value, **tolerances
                )
        # non-constant variables
        else:
            is_within_range = self.value_range[0] <= value <= self.value_range[1]
            is_within_tolerance = any(
                [math.isclose(value, ele, **tolerances) for ele in self.value_range]
            )
        return is_within_range or is_within_tolerance


class DistributionVariable(Variable):
    """Variable for distributions. Must be a subclass of torch.distributions.Distribution.

    Attributes:
        unit: Unit associated with the variable.
    """

    unit: Optional[str] = None

    @property
    def default_validation_config(self) -> ConfigEnum:
        return "none"

    def validate_value(self, value: TDistribution, config: ConfigEnum = None):
        """
        Validates the given value.

        Args:
            value (Distribution): The value to be validated.
            config (ConfigEnum, optional): The configuration for validation. Defaults to None.
              Allowed values are "none", "warn", and "error".

        Raises:
            TypeError: If the value is not an instance of Distribution.
        """
        _config = self.default_validation_config if config is None else config
        # mandatory validation
        self._validate_value_type(value)
        # optional validation
        if config != "none":
            pass  # not implemented

    @staticmethod
    def _validate_value_type(value: TDistribution):
        if not isinstance(value, TDistribution):
            raise TypeError(
                f"Expected value to be of type {TDistribution}, "
                f"but received {type(value)}."
            )


def get_variable(name: str) -> Type[Variable]:
    """Returns the Variable subclass with the given name.

    Args:
        name: Name of the Variable subclass.

    Returns:
        Variable subclass with the given name.
    """
    classes = [ScalarVariable, DistributionVariable]
    class_lookup = {c.__name__: c for c in classes}
    if name not in class_lookup.keys():
        raise KeyError(
            f"No variable named {name}, valid names are {list(class_lookup.keys())}"
        )
    return class_lookup[name]



import os
import logging
from typing import Union
from copy import deepcopy

import torch
from pydantic import field_validator
from botorch.models.transforms.input import ReversibleInputTransform

from lume_model.base import LUMEBaseModel
from lume_model.variables import ScalarVariable
from lume_model.models.utils import itemize_dict, format_inputs, InputDictModel

logger = logging.getLogger(__name__)


class TorchModel(LUMEBaseModel):
    """LUME-model class for torch models.

    By default, the models are assumed to be fixed, so all gradient computation is deactivated and the model and
    transformers are put in evaluation mode.

    Attributes:
        model: The torch base model.
        input_variables: List defining the input variables and their order.
        output_variables: List defining the output variables and their order.
        input_transformers: List of transformer objects to apply to input before passing to model.
        output_transformers: List of transformer objects to apply to output of model.
        output_format: Determines format of outputs: "tensor" or "raw".
        device: Device on which the model will be evaluated. Defaults to "cpu".
        fixed_model: If true, the model and transformers are put in evaluation mode and all gradient
          computation is deactivated.
        precision: Precision of the model, either "double" or "single".
    """

    model: torch.nn.Module
    input_transformers: list[Union[ReversibleInputTransform, torch.nn.Linear]] = None
    output_transformers: list[Union[ReversibleInputTransform, torch.nn.Linear]] = None
    output_format: str = "tensor"
    device: Union[torch.device, str] = "cpu"
    fixed_model: bool = True
    precision: str = "double"

    def __init__(self, *args, **kwargs):
        """Initializes TorchModel.

        Args:
            *args: Accepts a single argument which is the model configuration as dictionary, YAML or JSON
              formatted string or file path.
            **kwargs: See class attributes.
        """
        super().__init__(*args, **kwargs)
        self.input_transformers = (
            [] if self.input_transformers is None else self.input_transformers
        )
        self.output_transformers = (
            [] if self.output_transformers is None else self.output_transformers
        )

        # dtype property sets precision across model and transformers
        self.dtype

        # fixed model: set full model in eval mode and deactivate all gradients
        if self.fixed_model:
            is_scripted = isinstance(self.model, torch.jit.ScriptModule)
            self.model.eval().requires_grad_(False) if not is_scripted else None
            for t in self.input_transformers + self.output_transformers:
                if isinstance(t, torch.nn.Module):
                    t.eval().requires_grad_(False)

        # ensure consistent device
        self.to(self.device)

    @property
    def dtype(self):
        if self.precision == "double":
            self._dtype = torch.double
        elif self.precision == "single":
            self._dtype = torch.float
        else:
            raise ValueError(
                f"Unknown precision {self.precision}, "
                f"expected one of ['double', 'single']."
            )
        self._set_precision(self._dtype)
        return self._dtype

    @property
    def _tkwargs(self):
        return {"device": self.device, "dtype": self.dtype}

    @field_validator("model", mode="before")
    def validate_torch_model(cls, v):
        if isinstance(v, (str, os.PathLike)):
            if os.path.exists(v):
                fname = v
                try:
                    v = torch.jit.load(v)
                    print(f"Loaded TorchScript (JIT) model from file: {fname}")
                except RuntimeError:
                    v = torch.load(v, weights_only=False)
                    print(f"Loaded PyTorch model from file: {fname}")
            else:
                raise OSError(f"File {v} is not found.")
        return v

    @field_validator("input_variables")
    def verify_input_default_value(cls, value):
        """Verifies that input variables have the required default values."""
        for var in value:
            if var.default_value is None:
                raise ValueError(
                    f"Input variable {var.name} must have a default value."
                )
        return value

    @field_validator("input_transformers", "output_transformers", mode="before")
    def validate_transformers(cls, v):
        if not isinstance(v, list):
            raise ValueError("Transformers must be passed as list.")
        loaded_transformers = []
        for t in v:
            if isinstance(t, (str, os.PathLike)):
                if os.path.exists(t):
                    t = torch.load(t, weights_only=False)
                else:
                    raise OSError(f"File {t} is not found.")
            loaded_transformers.append(t)
        v = loaded_transformers
        return v

    @field_validator("output_format")
    def validate_output_format(cls, v):
        supported_formats = ["tensor", "variable", "raw"]
        if v not in supported_formats:
            raise ValueError(
                f"Unknown output format {v}, expected one of {supported_formats}."
            )
        return v

    def _set_precision(self, value: torch.dtype):
        """Sets the precision of the model."""
        self.model.to(dtype=value)
        for t in self.input_transformers + self.output_transformers:
            if isinstance(t, torch.nn.Module):
                t.to(dtype=value)

    def _evaluate(
        self,
        input_dict: dict[str, Union[float, torch.Tensor]],
    ) -> dict[str, Union[float, torch.Tensor]]:
        """Evaluates model on the given input dictionary.

        Args:
            input_dict: Input dictionary on which to evaluate the model.

        Returns:
            Dictionary of output variable names to values.
        """
        formatted_inputs = format_inputs(input_dict)
        input_tensor = self._arrange_inputs(formatted_inputs)
        input_tensor = self._transform_inputs(input_tensor)
        output_tensor = self.model(input_tensor)
        output_tensor = self._transform_outputs(output_tensor)
        parsed_outputs = self._parse_outputs(output_tensor)
        output_dict = self._prepare_outputs(parsed_outputs)
        return output_dict

    def input_validation(self, input_dict: dict[str, Union[float, torch.Tensor]]):
        """Validates input dictionary before evaluation.

        Args:
            input_dict: Input dictionary to validate.

        Returns:
            Validated input dictionary.
        """
        # validate input type (ints only are cast to floats for scalars)
        validated_input = InputDictModel(input_dict=input_dict).input_dict
        # format inputs as tensors w/o changing the dtype
        formatted_inputs = format_inputs(validated_input)
        # check default values for missing inputs
        filled_inputs = self._fill_default_inputs(formatted_inputs)
        # itemize inputs for validation
        itemized_inputs = itemize_dict(filled_inputs)

        for ele in itemized_inputs:
            # validate values that were in the torch tensor
            # any ints in the torch tensor will be cast to floats by Pydantic
            # but others will be caught, e.g. booleans
            ele = InputDictModel(input_dict=ele).input_dict
            # validate each value based on its var class and config
            super().input_validation(ele)

        # return the validated input dict for consistency w/ casting ints to floats
        if any([isinstance(value, torch.Tensor) for value in validated_input.values()]):
            validated_input = {
                k: v.to(**self._tkwargs) for k, v in validated_input.items()
            }

        return validated_input

    def output_validation(self, output_dict: dict[str, Union[float, torch.Tensor]]):
        """Itemizes tensors before performing output validation."""
        itemized_outputs = itemize_dict(output_dict)
        for ele in itemized_outputs:
            super().output_validation(ele)

    def random_input(self, n_samples: int = 1) -> dict[str, torch.Tensor]:
        """Generates random input(s) for the model.

        Args:
            n_samples: Number of random samples to generate.

        Returns:
            Dictionary of input variable names to tensors.
        """
        input_dict = {}
        for var in self.input_variables:
            if isinstance(var, ScalarVariable):
                input_dict[var.name] = var.value_range[0] + torch.rand(
                    size=(n_samples,)
                ) * (var.value_range[1] - var.value_range[0])
            else:
                torch.tensor(var.default_value, **self._tkwargs).repeat((n_samples, 1))
        return input_dict

    def random_evaluate(
        self, n_samples: int = 1
    ) -> dict[str, Union[float, torch.Tensor]]:
        """Returns random evaluation(s) of the model.

        Args:
            n_samples: Number of random samples to evaluate.

        Returns:
            Dictionary of variable names to outputs.
        """
        random_input = self.random_input(n_samples)
        return self.evaluate(random_input)

    def to(self, device: Union[torch.device, str]):
        """Updates the device for the model, transformers and default values.

        Args:
            device: Device on which the model will be evaluated.
        """
        self.model.to(device)
        for t in self.input_transformers + self.output_transformers:
            if isinstance(t, torch.nn.Module):
                t.to(device)
        self.device = device

    def insert_input_transformer(
        self, new_transformer: ReversibleInputTransform, loc: int
    ):
        """Inserts an additional input transformer at the given location.

        Args:
            new_transformer: New transformer to add.
            loc: Location where the new transformer shall be added to the transformer list.
        """
        self.input_transformers = (
            self.input_transformers[:loc]
            + [new_transformer]
            + self.input_transformers[loc:]
        )

    def insert_output_transformer(
        self, new_transformer: ReversibleInputTransform, loc: int
    ):
        """Inserts an additional output transformer at the given location.

        Args:
            new_transformer: New transformer to add.
            loc: Location where the new transformer shall be added to the transformer list.
        """
        self.output_transformers = (
            self.output_transformers[:loc]
            + [new_transformer]
            + self.output_transformers[loc:]
        )

    def update_input_variables_to_transformer(
        self, transformer_loc: int
    ) -> list[ScalarVariable]:
        """Returns input variables updated to the transformer at the given location.

        Updated are the value ranges and default of the input variables. This allows, e.g., to add a
        calibration transformer and to update the input variable specification accordingly.

        Args:
            transformer_loc: The location of the input transformer to adjust for.

        Returns:
            The updated input variables.
        """
        x_old = {
            "min": torch.tensor(
                [var.value_range[0] for var in self.input_variables], dtype=self.dtype
            ),
            "max": torch.tensor(
                [var.value_range[1] for var in self.input_variables], dtype=self.dtype
            ),
            "default": torch.tensor(
                [var.default_value for var in self.input_variables], dtype=self.dtype
            ),
        }
        x_new = {}
        for key in x_old.keys():
            x = x_old[key]

            # Make at least 2D
            if x.ndim == 0:
                x = x.unsqueeze(0)
            if x.ndim == 1:
                x = x.unsqueeze(0)

            # compute previous limits at transformer location
            for i in range(transformer_loc):
                if isinstance(self.input_transformers[i], ReversibleInputTransform):
                    x = self.input_transformers[i].transform(x)
                else:
                    x = self.input_transformers[i](x)
            # untransform of transformer to adjust for
            if isinstance(
                self.input_transformers[transformer_loc], ReversibleInputTransform
            ):
                x = self.input_transformers[transformer_loc].untransform(x)
            else:
                w = self.input_transformers[transformer_loc].weight
                b = self.input_transformers[transformer_loc].bias
                x = torch.matmul((x - b), torch.linalg.inv(w.T))
            # backtrack through transformers
            for transformer in self.input_transformers[:transformer_loc][::-1]:
                if isinstance(
                    self.input_transformers[transformer_loc], ReversibleInputTransform
                ):
                    x = transformer.untransform(x)
                else:
                    w, b = transformer.weight, transformer.bias
                    x = torch.matmul((x - b), torch.linalg.inv(w.T))
            x_new[key] = x
        updated_variables = deepcopy(self.input_variables)
        for i, var in enumerate(updated_variables):
            var.value_range = [x_new["min"][0][i].item(), x_new["max"][0][i].item()]
            var.default_value = x_new["default"][0][i].item()
        return updated_variables

    def _fill_default_inputs(
        self, input_dict: dict[str, torch.Tensor]
    ) -> dict[str, torch.Tensor]:
        """Fills missing input variables with default values.

        Args:
            input_dict: Dictionary of input variable names to tensors.

        Returns:
            Dictionary of input variable names to tensors with default values for missing inputs.
        """
        for var in self.input_variables:
            if var.name not in input_dict.keys():
                input_dict[var.name] = torch.tensor(var.default_value, **self._tkwargs)
        return input_dict

    def _arrange_inputs(
        self, formatted_inputs: dict[str, torch.Tensor]
    ) -> torch.Tensor:
        """Enforces order of input variables.

        Enforces the order of the input variables to be passed to the transformers and model and updates the
        returned tensor with default values for any inputs that are missing.

        Args:
            formatted_inputs: Dictionary of input variable names to tensors.

        Returns:
            Ordered input tensor to be passed to the transformers.
        """
        default_tensor = torch.tensor(
            [var.default_value for var in self.input_variables], **self._tkwargs
        )

        # determine input shape
        input_shapes = [formatted_inputs[k].shape for k in formatted_inputs.keys()]
        if not all(ele == input_shapes[0] for ele in input_shapes):
            raise ValueError("Inputs have inconsistent shapes.")

        input_tensor = torch.tile(default_tensor, dims=(*input_shapes[0], 1))
        for key, value in formatted_inputs.items():
            input_tensor[..., self.input_names.index(key)] = value

        if input_tensor.shape[-1] != len(self.input_names):
            raise ValueError(
                f"""
                Last dimension of input tensor doesn't match the expected number of inputs\n
                received: {default_tensor.shape}, expected {len(self.input_names)} as the last dimension
                """
            )
        return input_tensor

    def _transform_inputs(self, input_tensor: torch.Tensor) -> torch.Tensor:
        """Applies transformations to the inputs.

        Args:
            input_tensor: Ordered input tensor to be passed to the transformers.

        Returns:
            Tensor of transformed inputs to be passed to the model.
        """
        # Make at least 2D
        if input_tensor.ndim == 0:
            input_tensor = input_tensor.unsqueeze(0)
        if input_tensor.ndim == 1:
            input_tensor = input_tensor.unsqueeze(0)

        for transformer in self.input_transformers:
            if isinstance(transformer, ReversibleInputTransform):
                input_tensor = transformer.transform(input_tensor)
            else:
                input_tensor = transformer(input_tensor)
        return input_tensor

    def _transform_outputs(self, output_tensor: torch.Tensor) -> torch.Tensor:
        """(Un-)Transforms the model output tensor.

        Args:
            output_tensor: Output tensor from the model.

        Returns:
            (Un-)Transformed output tensor.
        """
        for transformer in self.output_transformers:
            if isinstance(transformer, ReversibleInputTransform):
                output_tensor = transformer.untransform(output_tensor)
            else:
                w, b = transformer.weight, transformer.bias
                output_tensor = torch.matmul((output_tensor - b), torch.linalg.inv(w.T))
        return output_tensor

    def _parse_outputs(self, output_tensor: torch.Tensor) -> dict[str, torch.Tensor]:
        """Constructs dictionary from model output tensor.

        Args:
            output_tensor: (Un-)transformed output tensor from the model.

        Returns:
            Dictionary of output variable names to (un-)transformed tensors.
        """
        parsed_outputs = {}
        if output_tensor.dim() in [0, 1]:
            output_tensor = output_tensor.unsqueeze(0)
        if len(self.output_names) == 1:
            parsed_outputs[self.output_names[0]] = output_tensor.squeeze()
        else:
            for idx, output_name in enumerate(self.output_names):
                parsed_outputs[output_name] = output_tensor[..., idx].squeeze()
        return parsed_outputs

    def _prepare_outputs(
        self,
        parsed_outputs: dict[str, torch.Tensor],
    ) -> dict[str, Union[float, torch.Tensor]]:
        """Updates and returns outputs according to output_format.

        Updates the output variables within the model to reflect the new values.

        Args:
            parsed_outputs: Dictionary of output variable names to transformed tensors.

        Returns:
            Dictionary of output variable names to values depending on output_format.
        """
        if self.output_format.lower() == "tensor":
            return parsed_outputs
        else:
            return {
                key: value.item() if value.squeeze().dim() == 0 else value
                for key, value in parsed_outputs.items()
            }



import os
import sys
import yaml
import importlib
from typing import Union, get_origin, get_args

from lume_model.variables import ScalarVariable, get_variable


def try_import_module(name: str):
    """Tries to import module if required.

    Args:
        name: Module name.

    Returns:
        Imported module if successful, None otherwise.
    """
    if name not in sys.modules:
        try:
            module = importlib.import_module(name)
        except ImportError:
            module = None
    else:
        module = sys.modules[name]
    return module


def verify_unique_variable_names(variables: list[ScalarVariable]):
    """Verifies that variable names are unique.

    Raises a ValueError if any reoccurring variable names are found.

    Args:
        variables: List of scalar variables.
    """
    names = [var.name for var in variables]
    non_unique_names = [name for name in set(names) if names.count(name) > 1]
    if non_unique_names:
        raise ValueError(f"Variable names {non_unique_names} are not unique.")


def serialize_variables(v: dict):
    """Performs custom serialization for in- and output variables.

    Args:
        v: Object to serialize.

    Returns:
        Dictionary with serialized in- and output variables.
    """
    for key, value in v.items():
        if key in ["input_variables", "output_variables"] and isinstance(value, list):
            v[key] = {
                var_dict["name"]: {
                    var_k: var_v
                    for var_k, var_v in var_dict.items()
                    if not (var_k == "name" or var_v is None)
                }
                for var_dict in value
            }
    return v


def deserialize_variables(v):
    """Performs custom deserialization for in- and output variables.

    Args:
        v: Object to deserialize.

    Returns:
        Dictionary with deserialized in- and output variables.
    """
    for key, value in v.items():
        if key in ["input_variables", "output_variables"] and isinstance(value, dict):
            v[key] = [
                var_dict | {"name": var_name} for var_name, var_dict in value.items()
            ]
    return v


def variables_as_yaml(
    input_variables: list[ScalarVariable],
    output_variables: list[ScalarVariable],
    file: Union[str, os.PathLike] = None,
) -> str:
    """Returns and optionally saves YAML formatted string defining the in- and output variables.

    Args:
        input_variables: List of input variables.
        output_variables: List of output variables.
        file: If not None, YAML formatted string is saved to given file path.

    Returns:
        YAML formatted string defining the in- and output variables.
    """
    for variables in [input_variables, output_variables]:
        verify_unique_variable_names(variables)
    v = {
        "input_variables": [var.model_dump() for var in input_variables],
        "output_variables": [var.model_dump() for var in output_variables],
    }
    s = yaml.safe_dump(serialize_variables(v), default_flow_style=None, sort_keys=False)
    if file is not None:
        with open(file, "w") as f:
            f.write(s)
    return s


def variables_from_dict(
    config: dict,
) -> tuple[list[ScalarVariable], list[ScalarVariable]]:
    """Parses given config and returns in- and output variable lists.

    Args:
        config: Variable configuration.

    Returns:
        In- and output variable lists.
    """
    input_variables, output_variables = [], []
    for key, value in {**config}.items():
        if key in ["input_variables", "output_variables"]:
            for var in value:
                variable_class = get_variable(var["variable_class"])
                if key == "input_variables":
                    input_variables.append(variable_class(**var))
                elif key == "output_variables":
                    output_variables.append(variable_class(**var))
    for variables in [input_variables, output_variables]:
        verify_unique_variable_names(variables)
    return input_variables, output_variables


def variables_from_yaml(
    yaml_obj: Union[str, os.PathLike],
) -> tuple[list[ScalarVariable], list[ScalarVariable]]:
    """Parses YAML object and returns in- and output variable lists.

    Args:
        yaml_obj: YAML formatted string or file path.

    Returns:
        In- and output variable lists.
    """
    if os.path.exists(yaml_obj):
        with open(yaml_obj) as f:
            yaml_str = f.read()
    else:
        yaml_str = yaml_obj
    config = deserialize_variables(yaml.safe_load(yaml_str))
    return variables_from_dict(config)


def get_valid_path(
    path: Union[str, os.PathLike],
    directory: Union[str, os.PathLike] = "",
) -> Union[str, os.PathLike]:
    """Validates path exists either as relative or absolute path and returns the first valid option.

    Args:
        path: Path to validate.
        directory: Directory against which relative paths are checked.

    Returns:
        The first valid path option as an absolute path.
    """
    relative_path = os.path.join(directory, path)
    if os.path.exists(relative_path):
        return os.path.abspath(relative_path)
    elif os.path.exists(path):
        return os.path.abspath(path)
    else:
        raise OSError(f"File {path} is not found.")


def replace_relative_paths(
    d: dict,
    model_fields: dict = None,
    directory: Union[str, os.PathLike] = "",
) -> dict:
    """Replaces dictionary entries with absolute paths where the model field annotation is not string or path-like.

    Args:
        d: Dictionary to process.
        model_fields: Model fields dictionary used to check expected type.
        directory: Directory against which relative paths are checked.

    Returns:
        Dictionary with replaced paths.
    """
    if model_fields is None:
        model_fields = {}
    for k, v in d.items():
        if isinstance(v, (str, os.PathLike)):
            if k in model_fields.keys():
                field_types = [model_fields[k].annotation]
                if get_origin(model_fields[k].annotation) is Union:
                    field_types = list(get_args(model_fields[k].annotation))
                if all([t not in field_types for t in [str, os.PathLike]]):
                    d[k] = get_valid_path(v, directory)
        elif isinstance(v, list):
            if k in model_fields.keys():
                field_types = []
                for i, field_type in enumerate(get_args(model_fields[k].annotation)):
                    if get_origin(field_type) is Union:
                        field_types.extend(list(get_args(field_type)))
                    else:
                        field_types.append(field_type)
                for i, ele in enumerate(v):
                    if isinstance(ele, (str, os.PathLike)) and all(
                        [t not in field_types for t in [str, os.PathLike]]
                    ):
                        v[i] = get_valid_path(ele, directory)
        elif isinstance(v, dict):
            model_subfields = {
                ".".join(key.split(".")[1:]): value
                for key, value in model_fields.items()
                if key.startswith(f"{k}.")
            }
            d[k] = replace_relative_paths(v, model_subfields, directory)
    return d



import os
import warnings
from typing import Dict, Any, Union

from torch import Tensor, nn

try:
    import mlflow

    HAS_MLFLOW = True
    import logging

    logger = logging.getLogger("mlflow")
    # Set log level to error until annoying signature warnings are fixed
    logger.setLevel(logging.ERROR)
except ImportError:
    pass


def register_model(
    lume_model,
    input: dict[str, Union[float, Tensor]] | Tensor,
    artifact_path: str,
    registered_model_name: str | None = None,
    tags: dict[str, Any] | None = None,
    version_tags: dict[str, Any] | None = None,
    alias: str | None = None,
    run_name: str | None = None,
    log_model_dump: bool = True,
    save_jit: bool = False,
    **kwargs,
) -> mlflow.models.model.ModelInfo:
    """
    Registers the model to MLflow if mlflow is installed. Each time this function is called, a new version
    of the model is created. The model is saved to the tracking server or local directory, depending on the
    MLFLOW_TRACKING_URI.

    If no tracking server is set up, data and artifacts are saved directly under your current directory. To set up
    a tracking server, set the environment variable MLFLOW_TRACKING_URI, e.g. a local port/path. See
    https://mlflow.org/docs/latest/getting-started/intro-quickstart/ for more info.

    Args:
        lume_model: LumeModel to register.
        input: Input dictionary to infer the model signature.
        artifact_path: Path to store the model in MLflow.
        registered_model_name: Name of the registered model in MLflow.
        tags: Tags to add to the MLflow model.
        version_tags: Tags to add to this MLflow model version.
        alias: Alias to add to this MLflow model version.
        run_name: Name of the MLflow run.
        log_model_dump: If True, the model dump is logged to MLflow.
        save_jit: If True, the model is saved as a JIT model when calling model.dump, if log_model_dump=True.
        **kwargs: Additional arguments for mlflow.pyfunc.log_model.

    Returns:
        Model info metadata, mlflow.models.model.ModelInfo.
    """
    if not HAS_MLFLOW:
        raise ImportError("MLflow is not installed. Cannot register model.")
    if "MLFLOW_TRACKING_URI" not in os.environ:
        warnings.warn(
            "MLFLOW_TRACKING_URI is not set. Data and artifacts will be saved directly under your current directory."
        )

    # Log the model to MLflow
    with mlflow.start_run(run_name=run_name):
        # Define the signature of the model
        if isinstance(lume_model, nn.Module):
            signature = mlflow.models.infer_signature(
                input.numpy(), lume_model(Tensor(input)).detach().numpy()
            )
            model_info = mlflow.pytorch.log_model(
                pytorch_model=lume_model,
                artifact_path=artifact_path,
                signature=signature,
                registered_model_name=registered_model_name,
                **kwargs,
            )
        else:
            # Create pyfunc model for MLflow to be able to log/load the model
            pf_model = create_mlflow_model(lume_model)
            # Adjust the input to match the expected input format
            # Must be one of `numpy.ndarray`, `List[numpy.ndarray]`, `Dict[str, numpy.ndarray]` or `pandas.DataFrame`
            input = {key: value.numpy() for key, value in input.items()}
            signature = mlflow.models.infer_signature(input, pf_model.predict(input))
            model_info = mlflow.pyfunc.log_model(
                python_model=pf_model,
                artifact_path=artifact_path,
                signature=signature,
                input_example=input,
                registered_model_name=registered_model_name,
                **kwargs,
            )

        if log_model_dump:
            # Log the model dump files to MLflow
            # TODO: pass directory where user wants local dump to, default to working directory
            run_name = mlflow.active_run().info.run_name
            name = registered_model_name or f"{run_name}"

            lume_model.dump(f"{name}.yml", save_jit=save_jit)
            mlflow.log_artifact(f"{name}.yml", artifact_path)
            mlflow.log_artifact(f"{name}_model.pt", artifact_path)
            os.remove(f"{name}.yml")
            os.remove(f"{name}_model.pt")
            if save_jit:
                mlflow.log_artifact(f"{name}_model.jit", artifact_path)
                os.remove(f"{name}_model.jit")

            # Get and log the input and output transformers
            lume_model = (
                lume_model._model if isinstance(lume_model, nn.Module) else lume_model
            )
            for i in range(len(lume_model.input_transformers)):
                mlflow.log_artifact(f"{name}_input_transformers_{i}.pt", artifact_path)
                os.remove(f"{name}_input_transformers_{i}.pt")
            for i in range(len(lume_model.output_transformers)):
                mlflow.log_artifact(f"{name}_output_transformers_{i}.pt", artifact_path)
                os.remove(f"{name}_output_transformers_{i}.pt")

    if (tags or alias or version_tags) and registered_model_name:
        from mlflow import MlflowClient

        client = MlflowClient()
        # Get the latest version of the registered model that we just registered
        latest_version = model_info.registered_model_version

        if tags:
            for key, value in tags.items():
                client.set_registered_model_tag(registered_model_name, key, value)
        if version_tags:
            for key, value in version_tags.items():
                client.set_model_version_tag(
                    registered_model_name, latest_version, key, value
                )
        if alias:
            client.set_registered_model_alias(
                registered_model_name, alias, latest_version
            )

    elif (tags or alias or version_tags) and not registered_model_name:
        warnings.warn(
            "No registered model name provided. Tags and aliases will not be set."
        )

    return model_info


def create_mlflow_model(model) -> mlflow.pyfunc.PythonModel:
    """Creates an MLflow model from the given model."""
    return PyFuncModel(model=model)


class PyFuncModel(mlflow.pyfunc.PythonModel):
    """
    Custom MLflow model class for LumeModel.
    Uses Pyfunc to define a model that can be saved and loaded with MLflow.

    Must implement the `predict` method.
    """

    def __init__(self, model):
        self.model = model

    def predict(self, model_input: Dict[str, Any]) -> Dict[str, Any]:
        """Evaluate the model with the given input."""
        # Convert input to the format expected by the model
        # TODO: this isn't very general but type validation in torch modules requires this. May need to adjust.
        model_input = {key: Tensor(value) for key, value in model_input.items()}
        return self.model.evaluate(model_input)

    def save_model(self):
        raise NotImplementedError("Save model not implemented")

    def load_model(self):
        raise NotImplementedError("Load model not implemented")

    def get_lume_model(self):
        return self.model



from typing import Union, Dict

from pydantic import BaseModel, ConfigDict
import torch
from torch.distributions import Distribution


def itemize_dict(
    d: dict[str, Union[float, torch.Tensor, Distribution]],
) -> list[dict[str, Union[float, torch.Tensor]]]:
    """Itemizes the given in-/output dictionary.

    Args:
        d: Dictionary to itemize.

    Returns:
        List of in-/output dictionaries, each containing only a single value per in-/output.
    """
    has_tensors = any([isinstance(value, torch.Tensor) for value in d.values()])
    itemized_dicts = []
    if has_tensors:
        for k, v in d.items():
            for i, ele in enumerate(v.flatten()):
                if i >= len(itemized_dicts):
                    itemized_dicts.append({k: ele.item()})
                else:
                    itemized_dicts[i][k] = ele.item()
    else:
        itemized_dicts = [d]
    return itemized_dicts


def format_inputs(
    input_dict: dict[str, Union[float, torch.Tensor]],
) -> dict[str, torch.Tensor]:
    """Formats values of the input dictionary as tensors.

    Args:
        input_dict: Dictionary of input variable names to values.

    Returns:
        Dictionary of input variable names to tensors.
    """
    formatted_inputs = {}
    for var_name, value in input_dict.items():
        v = value if isinstance(value, torch.Tensor) else torch.tensor(value)
        formatted_inputs[var_name] = v
    return formatted_inputs


class InputDictModel(BaseModel):
    """Pydantic model for input dictionary validation.

    Attributes:
        input_dict: Input dictionary to validate.
    """

    input_dict: Dict[str, Union[torch.Tensor, float]]

    model_config = ConfigDict(arbitrary_types_allowed=True, strict=True)
